function InstagramVideoControls() {

    let showLog = false;

    /** @var {HTMLVideoElement} */
    let player;

    let config = {};

    const LOCALSTORAGE_VOLUME = "controls_for_instagram_volume";

    // true after .init() was called.
    let initialized = false;

    let isInstagramDotCom;

    let mutationObserverEnabled = false;

    const knownVideoElements = new Set();

    /**
     *
     * @param {string} selector
     * @param {Element} [root]
     * @returns {Array<Element>}
     */
    function selectAll(selector, root) {
        return arr((root || document).querySelectorAll(selector));
    }

    function arr(a) {
        return Array.from(a);
    }

    function modifyVideos() {
        // log('modifyVideos()');

        // On 3rd party non-Instagram websites, they use a .EmbedVideo container to hold the video.
        // On Instagrams site, they don't.
        // So, we look for both styles.
        selectAll("video").forEach(video => {

            // Try to find a component root element, which is basically a div that's a few parents above the video element,
            // and serves as a container for all the instagram GUI stuff, like buttons, images, overlays, the video element etc...

            // .EmbedVideo is found when were in an iframe used by a 3rd party to embed instagram.
            let componentRoot = video.closest(".EmbedVideo");
            if (!componentRoot) {
                // Try to find the style of root used on instagram.com when they overlay the video with a big play button.
                componentRoot = getEstimatedVideoComponentRootElement(video);
            }

            if (!componentRoot) {
                // If we reached here, either they changed their dom structure, or its a video without the insta-junk markup. This seems to happen
                // on very long videos - they just present a native html5 player, although they still disable downloads and other stuff.
                // We'll just go up 4 parents like the other cases and hope for the best lol.
                // I think we aren't likely to find any overlay buttons in this case, at least I havent encountered any yet.
                componentRoot = nthParent(video, 4);

                // Also, we need to mark the video as not having a component root.
                video.dataset.icNoClickToPlayHandlerNeeded = "1";
            }

            if (componentRoot) {
                modifyVideoElement(componentRoot);
                removeOverlayWithInstagramPlayControl(componentRoot);
            }

        });
    }

    function hide(el) {
        el && el.style && (el.style.display = 'none');
    }

    /**
     * This element is their control which pauses the vid when you hold the mouse down.
     * It interferes with the <video> receiving mousemove events necessary to show the native controls,
     * and also interferes w/ normal click to toggle play/pause.
     *
     *
     * @param videoElement
     */
    function incapacitateStoryVideoPausingOverlays(videoElement) {
        // There's more than 1 element overlaying the video.
        // Also, they change this markup often.
        arr(videoElement.parentElement.children).filter(el => el !== videoElement).forEach(hide);
    }

    function modifyVideoElement(embedRootElem) {

        selectAll("video", embedRootElem).forEach(videoPlayer => {

            if (videoPlayer.dataset.nativeControlsISetJooFree) {
                // Skip, we've already processed this element.
                return;
            }

            // This is used as a css selector to redefine the style instatard uses to hide the native media controls.
            // We use the dataset instead of a class because the react framework that they use will overwrite some other properties, erasing our class name.
            videoPlayer.dataset.nativeControlsISetJooFree = "1";
            knownVideoElements.add(videoPlayer);

            // Enable native controls. Some instagram videos actually already have them enabled.
            videoPlayer.controls = true;

            // They tend to not show the download video option, so we reenable it by clearing the list of restricted controls.
            videoPlayer.setAttribute('controlsList', '');

            if (isInstagramStoriesPage()) {
                incapacitateStoryVideoPausingOverlays(videoPlayer);
            }

            // Setup a click handler that only fires once.
            const clickHandler = evt => {
                // We need to stop the event, otherwise clicking on the video element will open a new tab.
                evt.stopPropagation();

                // We only want to interfere on the first click, to start the video. It seems that after this point,
                // instagram installs its own click handlers that will play/pause the video, so we don't want to interfere and
                // end up playing their pause, or pausing their play.
                if (videoPlayer.currentTime === 0.0) {
                    videoPlayer.play();
                }

                log(`removing click handler from inside click handler`);
                // We also remove the event handler in case the video is rewound to 0.00 time, in which case an event loop can occur.
                videoPlayer.removeEventListener('click', clickHandler);
            };

            if (!videoPlayer.dataset.icNoClickToPlayHandlerNeeded) {
                videoPlayer.addEventListener('click', clickHandler);
            }

            videoPlayer.addEventListener('volumechange', evt => {
                if (config.rememberVolumeLevel && getSavedVolumeLevel() !== videoPlayer.volume) {
                    saveVolumeLevel(videoPlayer.volume);
                    setVolumeOfPreviouslySeenVideoElements(videoPlayer.volume);
                    log('The volume changed.', evt, videoPlayer.volume);
                }
            });

            // We set the volume immediately because this is the first time we've seen this element.
            const savedVolumeLevel = getSavedVolumeLevel();
            if (config.rememberVolumeLevel) {
                setVolumeIfChanged(videoPlayer, savedVolumeLevel);
            }

            setPlaybackRateIfChanged(videoPlayer, config.playbackRate);

            // We probably only need to add css to the page once, but im worried react might remove it in some obscure
            // case, so adding it over and over seems safer.
            redefineWebkitMediaControlHidingCssRule();
        });

        try {
            // This func has a high chance of failing w/ NPE if instagram changes their dom layout.
            // It's not critical for this function to succeed, so we just keep going if it throws.
            modifyVideoHeightIfSendMessageBoxIsBlockingVideoControls(embedRootElem);
        } catch (ex) {
            log(ex);
        }
    }

    function setPlaybackRateIfChanged(player, newPlaybackRate) {
        // Only update if the current playbackRate is different from the new playbackRate (avoids triggering a new ratechange event).
        if (player && !isNaN(newPlaybackRate) && valuesAreDifferentEnough(player.playbackRate,  newPlaybackRate)) {
            try {
                player.playbackRate = newPlaybackRate;
            } catch (ex) {
                log('failed to set playback rate. it was probably out of range.', {player, playbackRate: newPlaybackRate, ex});
            }
        }
    }

    function setVolumeIfChanged(player, newVolume) {
        // Only update if the current volume is different from the new volume (avoids triggering a new volumechange event).
        if (player && !isNaN(newVolume) && valuesAreDifferentEnough(player.volume,  newVolume)) {
            try {
                log('before set vol');
                player.volume = newVolume;
                // If you notice sometimes videos autoplay but start muted, it's likely this Chrome feature: https://developers.google.com/web/updates/2017/09/autoplay-policy-changes
            } catch (ex) {
                log('failed to set volume. it was probably out of range.', {player, volume: newVolume, ex});
            }
        }
    }

    function setVolumeOfPreviouslySeenVideoElements(volume) {
        getKnownVideoElements().forEach(video => {
            setVolumeIfChanged(video, volume);
        });
    }

    function setPlaybackRateOfPreviouslySeenVideoElements(playbackRate) {
        getKnownVideoElements().forEach(video => {
            setPlaybackRateIfChanged(video, playbackRate);
        });
    }

    /**
     *
     * @param {number} floatVal1
     * @param {number} floatVal2
     * @param {number} [minimumDiff]
     */
    function valuesAreDifferentEnough(floatVal1, floatVal2, minimumDiff = 0.01) {
        return Math.abs(floatVal1 > floatVal2 ? floatVal1 - floatVal2 : floatVal2 - floatVal1) >= minimumDiff;
    }

    function getKnownVideoElements() {
        return knownVideoElements;
    }

    function saveVolumeLevel(volume) {
        try {
            // We save to config as a backup, in case localStorage throws an exception.
            config.mostRecentVolume = volume;
            // @todo save to chrome storage instead of localStorage.
            window.localStorage.setItem(LOCALSTORAGE_VOLUME, volume);
        } catch (ex) {
            log('failed to save volume', ex);
        }
    }

    function getSavedVolumeLevel() {
        try {
            const strVal = window.localStorage.getItem(LOCALSTORAGE_VOLUME);
            const num = typeof strVal === 'string' ? Number(strVal) : 1;
            return isNaN(num) ? config.mostRecentVolume : num;
        } catch (ex) {
            log('failed to get saved volume', ex);
        }
    }

    /**
     * They set "display: none" to hide the native controls, so we redefine their oppressive style.
     */
    function redefineWebkitMediaControlHidingCssRule() {
        const css = `
       video[data-native-controls-i-set-joo-free]::-webkit-media-controls, video[data-nativeControlsISetJooFree]::-webkit-media-controls  {
            display: flex;
        }
        `;

        addCss(css);
    }

    function addCss(cssCode) {
        const styleElement = document.createElement("style");
        styleElement.appendChild(document.createTextNode(cssCode));
        document.getElementsByTagName("head")[0].appendChild(styleElement);
    }

    function removeOverlayWithInstagramPlayControl(embedRootElem) {
        // Remove the clear overlay that prevents click events from reaching the actual video element + controls.
        selectAll(".videoSpritePlayButton", embedRootElem).map(el => el.parentNode).forEach(hide);

        getVideoCoveringButtons(embedRootElem).forEach(hide);

        // They put an img element after the video element, which totally covers the video. Hide it.
        selectAll("video", embedRootElem).forEach(el => {
            // Might be nice to delay this so the video can load, or remove it only after the user
            // clicks - but then the click event wont hit the video element.
            arr(el.parentNode.childNodes).filter(el => el.tagName === 'IMG').forEach(hide);
        });
    }


    /**
     * On some stories videos, they overlay the video with a "send message" textarea, but it lays on top of the video
     * player controls, making them difficult to use. So, we will identify these videos, and when found, we will reduce
     * the height of the video, so it moves upwards in its parent box, allowing the controls to be accessed. The
     * send message doesnt move - it stays fixed to the bottom of the parent.
     */
    function modifyVideoHeightIfSendMessageBoxIsBlockingVideoControls(embedRootElem) {
        // Get the textarea.
        // We need to take a few steps back up the tree before we will reach a position that can search for the textarea.
        const textarea = nthParent(embedRootElem, 4).querySelector("textarea[placeholder='Send Message']");
        if (!textarea) {
            // log('no textarea');
            return;
        }

        // This is a div that holds the "send Message" textarea and button, but it does not contain the emoticons.
        const sendMessageContainer = nthParent(textarea, 4);
        if (!sendMessageContainer) {
            log('no sendMessageContainer');
            return;
        }

        // This contains the video, the Send Message stuff, and many buttons.
        const videoAndControlsContainer = nthParent(textarea, 6);
        if (!videoAndControlsContainer) {
            log('no videoAndControlsContainer');
            return;
        }

        // Make sure they didn't change dom layout.
        const numButtons = selectAll('button', videoAndControlsContainer).length
        if (numButtons < 2) {
            log('numButtons < 2 for videoAndControlsContainer');
            return;
        }

        // We measure the height of the textarea ancestor element, because its a bit larger and contains other elements, such as the submit button.
        const distanceFromTopOfSendMessageToBottomOfParentContainer = videoAndControlsContainer.getBoundingClientRect().bottom - sendMessageContainer.getBoundingClientRect().top;
        // We move a tiny bit more to make a nice looking gap. Not sure how reliable this nudge is, but whatever.
        const videoHeightAdjustment = distanceFromTopOfSendMessageToBottomOfParentContainer + 2;

        // Now, we change the height of the video, reducing it by the height of the problematic textarea box assembly.
        // The video should move upwards, leaving a black gap below it.
        const video = embedRootElem.querySelector("video");
        video.style.height = `calc(100% - ${videoHeightAdjustment}px)`;

        log(`video adjusted ${videoHeightAdjustment}px`);
    }

    // Should only be 1 button returned, if any, but that might change.
    function getVideoCoveringButtons(searchRoot) {
        // The button will always be very large - much bigger than a normal button since it needs to be the size of the video.
        return selectAll("[role=button]", searchRoot).filter(el => {
            const style = window.getComputedStyle(el);
            return style.width >= '200px' && style.height >= '200px';
        });
    }

    function getEstimatedVideoComponentRootElements(searchRoot) {
        // todo, use: video:not([data-native-controls-i-set-joo-free]) for efficiency.
        // but need to refactor to make sure overlays get removed from both this and mut obs code paths.
        // also consider if size was very small at first, causing overlay to not get detected at first, but then got resized - it should get removed, ideally.
        return selectAll("video", searchRoot).map(getEstimatedVideoComponentRootElement).filter(el => el);
    }

    /**
     * If you look in the dom, and traverse upwards from the video element, you'll soon find
     * a node which is the ancestor to both the video, and a div with role=button, and that div is big, covering the entire video.
     *
     * This func searches for that parent element. We use these as a sort of container element for the video and its
     * other controller pieces.
     *
     * @param {Element} video
     * @return {Element}
     */
    function getEstimatedVideoComponentRootElement(video) {
        for (const parent of nParents(video, 4)) {
            if (getVideoCoveringButtons(parent).length) {
                return parent;
            }
        }
        log("couldnt find parent", video);
    }

    /**
     * Returns N number of parents, ordered with the closest elements first.
     *
     * @param {Element} elem
     * @param {number} numParents
     * @return Element[]
     */
    function nParents(elem, numParents) {
        const parents = [];
        while (elem && --numParents >= 0) {
            elem = elem.parentElement;
            elem && parents.push(elem);
        }
        return parents;
    }

    function nthParent(elem, n) {
        while (elem && --n >= 0) {
            elem = elem.parentElement;
        }
        return elem;
    }

    function remove(el) {
        el.parentNode.removeChild(el);
    }

    function attachPageVisibilityListener() {
        document.addEventListener('visibilitychange', startLookingForPlayerDomElement);
    }

    function startLookingForPlayerDomElement() {
        // log("startLookingForPlayerDomElement");

        // We only look for the player when the page is visible, which
        // saves battery by not running in background/idle tabs.
        if (document.hidden) {
            return;
        }

        // Normally, on non-instagram.com pages, we only look for 1 video element because of instagram embeds will always
        // have just 1 video in the iframe. todo- We should probably look again after a dom modification event.
        // But, on instagram.com, they do a modal dialog that dynamically adds more, so we continually look for more videos.
        if (isInstagramDotCom) {

            if (isInstagramStoriesPage()) {
                // The stories GUI switches video elements very rapidly, and the videos are short duration.
                // They swap out video elements rapidly, so we need to react fast to changes, and apply our modifications
                // to each new video elements as they appear.
                // We use a MutationObserver to detect the new video elements.
                ensureMutationObserverEnabled();
                modifyVideos();
            } else {
                modifyVideos();
            }

            // We look more rapidly on instagram because the modal dynamically creates video elements.
            setTimeout(startLookingForPlayerDomElement, 2000);
        } else {
            if (player) {
                return;
            }

            modifyVideos();

            if (player) {
                log(player);
            } else {
                setTimeout(startLookingForPlayerDomElement, 5000);
            }
        }

    }

    function ensureMutationObserverEnabled() {
        // Prevent multiple observers.
        if (mutationObserverEnabled) {
            return;
        } else {
            mutationObserverEnabled = true;
        }

        function callback(mutations) {
            log(`mutations.length = ${mutations.length}`);
            mutations.forEach(mutation => {
                log(`mutation.addedNodes.length = ${mutation.addedNodes.length}`);
            });

            const videosAdded = [];
            mutations.forEach(mutation => {
                mutation.addedNodes.forEach(node => {
                    const videos = selectAll('video', node.parentElement);
                    videosAdded.push(... videos);
                });
            });

            log(`videosAdded.length = ${videosAdded.length}`, videosAdded);
            videosAdded.forEach(video => {
                // We pass the parent node as a cheap hack to allow reuse of the modifyVideoElement func
                modifyVideoElement(video.parentNode);
                // todo, maybe remove some other elements.
            });
        }

        let observer = new MutationObserver(callback);
        observer.observe(document.body, {subtree: true, childList: true});
    }

    function isInstagramStoriesPage() {
        return window.location.pathname.toLowerCase().startsWith('/stories/');
    }

    function isInstagramHostname() {
        return window.location.hostname.toLowerCase().endsWith('instagram.com');
    }

    function log() {
        if (!showLog) {
            return;
        }
        // const args = [location.href].concat(arguments);
        // const args = [location.href].concat(...arguments);
        console.log.apply(console, arguments);
    }

    this.setConfig = function(cfg) {
        // We need to be careful not to allow a 2nd call to setConfig() to cause updates to the volume of videos.
        // This could happen w/ multiple tabs open, where a user turns the vol down in 1 tab, which then gets saved to
        // chrome.storage, which then causes a storage event, which then cause the 2nd tab to receive the event and
        // call this setConfig() method, which would then update the volume of the videos. That would be annoying,
        // and prevent them from setting volume independently in 1 tab.
        // Note though, that if the page reloads, which is common, then the new volume setting from that other tab will be in effect.
        // hmmm... might be confusing behavior.
        log("new cfg", cfg);

        // Force the string to a number.
        cfg.playbackRate = Number(cfg.playbackRate);

        config = cfg;
        if (!initialized) {
            this.init();
            initialized = true;
        }

        setPlaybackRateOfPreviouslySeenVideoElements(config.playbackRate);
    };

    this.init = function() {
        isInstagramDotCom = isInstagramHostname();
        attachPageVisibilityListener();
        startLookingForPlayerDomElement();
        redefineWebkitMediaControlHidingCssRule();
    };

}

const ivc = new InstagramVideoControls();
window.instagramVideoControls = ivc;